#!/usr/bin/python3
#
# Copyright (c) Daniel Sheffield 2021
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
import time
import itertools
import sys
import urwid
import pandas as pd
from db_utils import QueryManager
from widgets import (
    NoTabCheckBox,
    AutoCompleteEdit,
    AutoCompleteFloatEdit,
    TransactionEditor,
    SuggestionPopup,
    GroceryTransactionEditor,
    ActivityManager,
    _set_focus_path,
)

try:
    from db_credentials import HOST, PASSWORD
    host = f'host={HOST}'
    password = f'password={PASSWORD}'
except:
    host = ''
    password = ''

try:
    import psycopg2
    from psycopg2.sql import SQL
    import os
    user = os.getenv('USER')
    conn = psycopg2.connect(f"{host} dbname=das user={user} {password}")
    cur = conn.cursor()
except:
    print('Failed to set up db connection. Entering Mock mode')
    from mock import *

palette = [
    ('banner', 'light gray', 'dark red'),
    ('streak', 'light red', 'dark gray'),
    ('bg', 'light red', 'black'),
]

grid_layout = [
    [ 'ts',          'store',    ],
    [ 'organic',     'product',  ],
    [ 'category',    'group',   ],
]
side_pane = [
    'unit',
    'quantity',
    'price',
]
bottom_pane = [
    'description',
    'dbview',
]

cols = [
    c for c in filter(
        lambda x: x is not None,
        itertools.chain(
            *grid_layout, side_pane, set(bottom_pane) - set(['dbview'])
        )
    )
]

display_map = {
    'ts': lambda x: f"{time.strftime('%Y-%m-%d %H:%M', (x.year, x.month, x.day, x.hour, x.minute, 0, 0, 0, 0))}",
    'price': lambda x: f"{x:.2f}",
    'quantity': lambda x: f"{x:.2f}",
    'organic': lambda x: "true" if x else "false",
}
display = lambda data, name: display_map[name](data) if name in display_map else data

def show_or_exit(key):
    if isinstance(key, tuple):
        return
    
    if key in ('esc',):
        raise urwid.ExitMainLoop()

def _apply_choice_callback(activity_manager, name, widget, value):
    txn = activity_manager.get('transaction')
    txn.apply_choice(name)(widget, value)
    activity_manager.show(txn.update())

def _show_suggestions_callback(activity_manager, name, options):
    txn = activity_manager.get('transaction')
    activity_manager.show(
        activity_manager.create(SuggestionPopup, None,
            txn.original_widget, name, options,
            lambda w,x: _apply_choice_callback(activity_manager, name, w, x),
            lambda: activity_manager.show(txn.update())
        ))

def _autocomplete_callback(activity_manager, query_manager, name, data):
    options = query_manager.unique_suggestions(name, **data)
    if len(options) > 0:
        _show_suggestions_callback(activity_manager, name, options)

def _save_and_clear_callback(activity_manager):
    txn = activity_manager.get('transaction')
    activity_manager.app.save(txn.data)
    txn.clear()
    activity_manager.show(txn.update())

args = sys.argv
log = args[1]

cur.execute("BEGIN")

activity_manager = ActivityManager()
query_manager = QueryManager(activity_manager, cur, display)

activity_manager.create(TransactionEditor, 'transaction',
    query_manager, cols, grid_layout, side_pane, bottom_pane,
    lambda: _save_and_clear_callback(activity_manager),
    lambda name, data: _autocomplete_callback(activity_manager, query_manager, name, data))

app = GroceryTransactionEditor(activity_manager, cur, log)

loop = urwid.MainLoop(app, palette, unhandled_input=show_or_exit)
loop.run()

cur.close()
conn.close()
