#!/usr/bin/python3
#
# Copyright (c) Daniel Sheffield 2021
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
import time
import itertools
import sys
import urwid
import pandas as pd
import numpy as np
from widgets import COPYRIGHT
from db_utils import QueryManager
from dateutil.parser import parse as parse_time
from widgets import (
    NoTabCheckBox,
    AutoCompleteEdit,
    AutoCompleteFloatEdit,
    SuggestionPopup,
    ActivityManager,
    #_set_focus_path,
)
from decimal import Decimal, InvalidOperation
from collections import (
    OrderedDict,
)

try:
    from db_credentials import HOST, PASSWORD
    host = f'host={HOST}'
    password = f'password={PASSWORD}'
except:
    host = ''
    password = ''

try:
    import psycopg2
    from psycopg2.sql import SQL
    import os
    user = os.getenv('USER')
    conn = psycopg2.connect(f"{host} dbname=das user={user} {password}")
    cur = conn.cursor()
except:
    print('Failed to set up db connection. Entering Mock mode')
    exit(1)
    #from mock import *

palette = [
    ('banner', 'light gray', 'dark red'),
    ('streak', 'light gray', 'black'),
    ('bg', 'dark red', 'black'),
]

top_pane = [
    'clear',
    'exit',
    'sort',
]

left_pane = [
    'product',
    'organic',
    'div',
    'spread',
    'rating',
]
right_pane = [
    'unit',
    'quantity',
    'price',
]
bottom_pane = [
    'dbview',
]

inputs = filter(
    lambda x: x is not None and x not in ('rating','spread', 'div'),
    itertools.chain(
        left_pane, right_pane
    )
)

display_map = {
    'price': lambda x: f"{x:.4f}",
    'quantity': lambda x: f"{x:.2f}",
    'organic': lambda x: "true" if x else "false",
}
display = lambda data, name: display_map[name](data) if name in display_map else data

def show_or_exit(key):
    if isinstance(key, tuple):
        return
    
    if key in ('esc',):
        raise urwid.ExitMainLoop()

def _apply_choice_callback(activity_manager, name, widget, value):
    activity = activity_manager.get('price_check')
    activity.apply_choice(name, widget, value)
    activity_manager.show(activity.update())

def _show_suggestions_callback(activity_manager, name, options):
    txn = activity_manager.get('price_check')
    activity_manager.show(
        activity_manager.create(SuggestionPopup, None,
            txn.original_widget, name, options,
            lambda w,x: _apply_choice_callback(activity_manager, name, w, x),
            lambda: activity_manager.show(txn.update())
        ))

def _autocomplete_callback(activity_manager, query_manager, name, data):
    options = query_manager.unique_suggestions(name, **data)
    if len(options) > 0:
        _show_suggestions_callback(activity_manager, name, options)

class GroceryPriceCheck(urwid.WidgetPlaceholder):
    def __init__(self, activity_manager):
        super().__init__(urwid.SolidFill(u'/'))
        self.activity_manager = activity_manager
        price_check = self.activity_manager.get('price_check')

        self.activity_manager.show(self)
        self.activity_manager.show(price_check)

class PriceCheck(urwid.WidgetPlaceholder):

    def iter_focus_paths(self):
        initial = [2,0,0,0]
        container = self.original_widget.original_widget.original_widget
        _set_focus_path(container, initial)
        while True:
            path = container.get_focus_path()
            yield path
            self.advance_focus()
            path = container.get_focus_path()
            if path == initial:
                self.advance_focus()
                break

    def advance_focus(self, reverse=False):
        container = self.original_widget.original_widget.original_widget

        path = container.get_focus_path()
        if reverse:
            paths = [ i for i in self.iter_focus_paths() ]
            zipped_paths = zip(paths, [
                *paths[1:], paths[0]
            ])
            prev_path = map(lambda x: x[0], filter(
                lambda x: x[1] == path,
                zipped_paths
            ))
            _set_focus_path(container, next(prev_path))
            return
        
        _iter = [ i for i in enumerate(path) ][::-1]

        for idx, part in _iter:
            p = [ i for i in path ]
            if reverse:
                p[idx] -= 1
            else:
                p[idx] += 1
            
            try:
                _set_focus_path(container, p)
                if path == [3]:
                    self.advance_focus(reverse=reverse)
                return
            except IndexError:
                path[idx] = 0
        
        container.set_focus_path([2,0,0,0])

    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        
        if getattr(self.original_widget.original_widget, 'original_widget', None) is None:
            return super().keypress(size, key)
        
        if key == 'tab':
            self.advance_focus()
        elif key == 'shift tab':
            self.advance_focus(reverse=True)
        else:
            return super().keypress(size, key)

    def apply_choice(self, name, widget, value):
        self.apply_changes(name, widget, value)
        for k,v in self.data.items():
            if k == name or v:
                continue
            options = self.query_manager.unique_suggestions(k, **self.data)
            if len(options) == 1:
                self.apply_changes(k, widget, list(options)[0])

    def apply_changes(self, name, widget, value):
        if name in self.data:
            
            self.data.update({
                name: value,
            })
        self.update_historic_prices()

    def clear(self):
        for k in self.data:
            if k == 'organic':
                self.data[k] = 'mixed'
                continue
            self.data[k] = ''
        self.update()
        return self
    
    def update(self):
        for k in self.edit_fields:
            if self.data[k] != self.edit_fields[k].get_edit_text():
                self.edit_fields[k].set_edit_text(self.data[k])
        
        if self.data['organic'] != self.organic_checkbox.state:
            self.organic_checkbox.set_state(self.data['organic'])
        
        self.update_historic_prices()
        
        return self
    
    def update_rating(self, _avg, _min, _max, price=None, quantity=None):
        if None in (_avg, _min, _max):
            return
        current = None if None in (price, quantity) else float(price/quantity)
        size = 16
        chars = ['|', *['-']*(size - 2), '|' ]
        rating = [' ']*len(chars)
        _min, _max = min(_min, current or _min), max(_max, current or _max)
        ls = np.linspace(_min, _max, len(chars))

        for idx, (e, a) in enumerate(zip(ls, ls[1:])):
            if e <= _avg < a:
                chars[max(idx-4,0)] = '['
                chars[min(idx+4,len(chars)-1)] = ']'
                for _idx, c in enumerate(f'{_avg:>7.2f}'):
                    chars[max(idx-3+_idx,0)] = c
            
            if current is not None and e <= current < a:
                rating[idx] = '^'
        
        self.text_fields['spread'].set_text(f"{_min:>7.2f}{''.join(chars)}{_max:<7.2f}")
        self.text_fields['rating'].set_text(f"{' '*7}{''.join(rating)}{' '*7}")
        
    
    def update_historic_prices(self):
        organic = None if self.data['organic'] == 'mixed' else self.data['organic']
        sort = '$/unit' if self.buttons['sort_price'].state else 'ts'
        product, unit = self.data['product'] or None, self.data['unit'] or None
        try: 
            price = Decimal(self.data['price']) or None
        except InvalidOperation:
            price = None
        
        try:
            quantity = Decimal(self.data['quantity']) or None
        except InvalidOperation:
            price = None
        
        if None not in (sort, product, unit):
            self.text_fields['dbview'].set_text(
                self.query_manager.get_historic_prices(
                    lambda *args: self.update_rating(*args, price=price, quantity=quantity),
                    sort, product, unit, organic=organic) 
            )

    def __init__(self, query_manager, fields,
        top_pane, left_pane, right_pane, bottom_pane,
        autocomplete_cb):
        super().__init__(urwid.SolidFill(u'/'))
        self.query_manager = query_manager
        self.top_pane = top_pane
        self.left_pane = left_pane
        self.right_pane = right_pane
        self.bottom_pane = bottom_pane
        self.edit_fields = OrderedDict()
        self.text_fields = OrderedDict()
        self.data = OrderedDict()
        self.organic_checkbox = NoTabCheckBox(
            u"Organic",
            state='mixed',
        )
        urwid.connect_signal(self.organic_checkbox, 'change', lambda w,v: self.apply_changes('organic', w, v))
        
        fields = [f for f in fields ]
        for k in fields:
            if k in self.right_pane and k != 'unit':
                ef = AutoCompleteFloatEdit(('bg', k), apply_change_func=lambda name: autocomplete_cb(name, self.data))
            elif k != 'organic':
                ef = AutoCompleteEdit(('bg', k), apply_change_func=lambda name: autocomplete_cb(name, self.data))
            elif k == 'organic':
                self.data[k] = 'mixed'
                continue
            
            self.data[k] = ''
            urwid.connect_signal(ef, 'change', lambda w,v,name=k: self.apply_changes(name, w, v))
            self.edit_fields[k] = ef
        
        self.buttons = OrderedDict()
        group = []
        for k in top_pane:
            if k != 'sort':
                self.buttons.update({ k:  urwid.Button(('streak', f'{k.title()}')) })
                continue
            self.buttons.update({ f'{k}_price':  urwid.RadioButton(group, ('streak', u'Best'), state="first True") })
            self.buttons.update({ f'{k}_date':  urwid.RadioButton(group, ('streak', u'Last'), state="first True") })
            for button in group:
                urwid.connect_signal(button, 'postchange', lambda *args: self.update_historic_prices())
        
        urwid.connect_signal(self.buttons['clear'], 'click', lambda x: self.clear().update())
        urwid.connect_signal(self.buttons['exit'], 'click', lambda x: show_or_exit('esc'))
        
        dbview, spread, rating = [ urwid.Text('') for i in range(0,3) ]
        
        self.text_fields.update({
            'dbview': dbview,
            'spread': spread,
            'rating': rating,
        })
        
        self.clear()
        
        header = urwid.Text(u'Price Check', 'center')
        _copyright = urwid.Text(COPYRIGHT, 'center')
        
        banner = urwid.Pile([
            urwid.Padding(header, 'center', width=('relative', 100)),
            urwid.Padding(_copyright, 'center', width=('relative', 100)),
        ])
        banner = urwid.AttrMap(banner, 'banner')
        fields = dict([
            (k, urwid.LineBox(urwid.AttrMap(self.edit_fields[k], 'streak'), title=k.title(), title_align='left')) for k in self.edit_fields
        ])
        fields['organic'] = urwid.LineBox(urwid.AttrMap(self.organic_checkbox, 'bg'))
        fields.update({
            'dbview': urwid.LineBox(
                urwid.AttrMap(dbview, 'streak'),
                title="Historic Prices",
                title_align='center',
            ),
            'div': urwid.Divider(),
            'spread': spread,
            'rating': rating,
        })

        right_pane_widget = (16, urwid.Pile(map(
            lambda x: fields[x] if x is not None else urwid.Divider(),
            self.right_pane
        )))
        left_pane_widget = (urwid.Pile(map(
            lambda x: fields[x] if x is not None else urwid.Divider(),
            self.left_pane
        )))
        
        widget = urwid.Pile([
            banner,
            urwid.Divider(),
            urwid.Columns(
                [
                    (9, urwid.Pile([
                        urwid.Divider(),
                        urwid.AttrMap(self.buttons['clear'], 'streak'),
                        urwid.Divider(),
                    ])),
                    urwid.LineBox(
                        urwid.Columns([ v for k,v in self.buttons.items() if 'sort' in k]),
                        title="Sort price by",
                        title_align='left',
                    ),
                    (9, urwid.Pile([
                        urwid.Divider(),
                        urwid.AttrMap(self.buttons['exit'], 'streak'),
                        urwid.Divider(),
                    ]))
                ],
                dividechars=2,
            ),
            urwid.Columns((left_pane_widget, right_pane_widget),
                dividechars=2,
            ),
            *[ fields[c] if c is not None else urwid.Divider() for c in self.bottom_pane ],
        ])
        widget = urwid.Filler(widget, 'top') 
        widget = urwid.AttrMap(widget, 'bg')
        self.original_widget = widget

cur.execute("BEGIN")

activity_manager = ActivityManager()
query_manager = QueryManager(activity_manager, cur, display)

activity_manager.create(PriceCheck, 'price_check',
    query_manager, inputs, top_pane, left_pane, right_pane, bottom_pane,
    lambda name, data: _autocomplete_callback(activity_manager, query_manager, name, data))

app = GroceryPriceCheck(activity_manager)

loop = urwid.MainLoop(app, palette, unhandled_input=show_or_exit)
loop.run()

cur.close()
conn.close()
